{
  "type": "record",
  "name": "DomainProperties",
  "namespace": "com.linkedin.pegasus2avro.domain",
  "doc": "Information about a Domain",
  "fields": [
    {
      "name": "name",
      "type": "string",
      "doc": "Display name of the Domain",
      "Searchable": {
        "boostScore": 10.0,
        "enableAutocomplete": true,
        "fieldType": "TEXT_PARTIAL"
      }
    },
    {
      "name": "description",
      "type": [
        "null",
        "string"
      ],
      "doc": "Description of the Domain",
      "default": null
    }
  ],
  "Aspect": {
    "name": "domainProperties"
  }
}