{
  "type": "record",
  "name": "DataHubUpgradeKey",
  "namespace": "com.linkedin.pegasus2avro.metadata.key",
  "doc": "Key for a DataHubUpgrade",
  "fields": [
    {
      "name": "id",
      "type": "string"
    }
  ],
  "Aspect": {
    "name": "dataHubUpgradeKey"
  }
}